#pragma once

#include <cstdio>
#include <cuda.h>
#include <cuda_runtime_api.h>
#include <cuda_runtime.h>
#include <cufft.h>
#include <math.h>
#include "GPUerror.h"

int const TpB = 512;		// Threads per Block

//==============================================================
// calculates hA[r] and hB[r] stored sequentially in h[2*M]
//--------------------------------------------------------------
__global__ void prepare_h(double *h, double *w_gpu, const int N, const int M)
{
         double *wm=w_gpu, *wp=w_gpu+M;
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= 2*M) return;
	if (tid < M) h[tid] = exp(-(wm[tid]+wp[tid])/N);
	else h[tid] = exp(-(-wm[tid-M]+wp[tid-M])/N);
}
//==============================================================
// normalizes the sums for phiA[r] and phiB[r]
//--------------------------------------------------------------
__global__ void normalize_phi(double *w_gpu, double *h, const double Q, const double N, const int M)
{
         double *phiA=h, *phiB=h+M, *hA=h, *hB=h+M;
         double *phim=w_gpu+2*M, *phip=w_gpu+3*M;
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= M) return;
	phiA[tid] = phim[tid]/(N*Q*hA[tid]);
	phiB[tid] = phip[tid]/(N*Q*hB[tid]);
	phim[tid] = phiA[tid] - phiB[tid];
	phip[tid] = phiA[tid] + phiB[tid];
}
//==============================================================
// performs the sums for phiA[r] and phiB[r]
//--------------------------------------------------------------
__global__ void sum_phi(double *phi_gpu, double *q1_gpu, double *q2_gpu, const int M)
{
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= M) return;
	phi_gpu[tid] += q1_gpu[tid]*q2_gpu[tid];
}
//==============================================================
// initializes array a with the value of b
//--------------------------------------------------------------
__global__ void Array_init(double *a, const double b, int const M)
{
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= M) return;
	a[tid] = b;
}
//==============================================================
// initializes array a with array b
//--------------------------------------------------------------
__global__ void Array_copy(double *a, const double *b, int const M)
{
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;
	if (tid >= M) return;
	a[tid] = b[tid];
}
//==============================================================
// multiplies complex array a with array b element-by-element
//--------------------------------------------------------------
__global__ void Mult_self(cufftDoubleComplex *a, const double *b, int const M)
{
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= M) return;
	a[tid].x *= b[tid];
	a[tid].y *= b[tid];
}
//==============================================================
// multiplies array a with array b element-by-element
//--------------------------------------------------------------
__global__ void Mult_self(double *a, const double *b, int const M)
{
	int const tid = threadIdx.x + blockIdx.x * blockDim.x;

	if (tid >= M) return;
	a[tid] *= b[tid];
}
