// fts.cc
//------------------------------------------------------------
// CPU version of the FTS code for a diblock copolymer melt
// Note that lengths are expressed in units of R0=a*N^0.5
//------------------------------------------------------------

#include<math.h>    // math subroutines
#include<stdlib.h>  // standard library
#include<fftw3.h>   // required for fast Fourier transforms
#include<time.h>    // required to seed random number generator
#include<complex>   // for complex-valued variables
#include"random.h"  // generates Gaussian random numbers 
#include"sim_eq.h"  // solves simultaneous linear equations
#define DIM 21      // max histories in Anderson mixing = DIM-1
using namespace std;

//------------------------------------------------------------
// Global variables:
//
// m[i] = number of mesh points in i'th dimension
// M = m[0]*m[1]*m[2] = total number of mesh points
// Mk = m[0]*m[1]*(m[2]/2+1) = total mesh points in k-space
// NA and NB = number of monomers in A and B blocks
// qr[r] = propagator in step subroutine
// qk[k] = Fourier transform (FT) of qr[r] times M/V
// wk[k] = FT of w-[r] times N*M/V
// g[k] = FT of Boltzmann weight for bond potential divided by M 
// qr_to_qk = plan for forward transform from qr[r] to qk[k] 
// qk_to_qr = plan for forward transform from qk[r] to qr[k] 
// wr_to_wk = plan for forward transform from w-[r] to wk[k] 

int    m[3], M, Mk, NA, NB;
double two_pi=8*atan(1.0), *qr, *g;
complex<double> *qk;
fftw_plan qr_to_qk, qk_to_qr, wr_to_wk;

//============================================================
// iterates the propagator: q_in=q_i(r) and q_out=q_{i+1}(r)
//------------------------------------------------------------
void step (double *q_in, double *q_out, double *g, double *h)
{
  for (int r=0; r<M; r++) qr[r] = q_in[r];
  fftw_execute(qr_to_qk);
  for (int k=0; k<Mk; k++) qk[k] *= g[k];
  fftw_execute(qk_to_qr);
  for (int r=0; r<M; r++) q_out[r] = qr[r]*h[r];
}
//============================================================
// evaluates phi- and phi+ stored at w+2*M and w+3*M
//------------------------------------------------------------
void diblock (double *w, double *lnQ)
{
  int    r, i, N=NA+NB;
  double q1[N+1][M], q2[N+1][M], Q, S1, S2;
  double phiA, phiB, hA[M], hB[M];

  for (r=0; r<M; r++) {
    hA[r] = exp(-(w[r+M]+w[r])/N); // Boltzmann weight for wA
    hB[r] = exp(-(w[r+M]-w[r])/N); // Boltzmann weight for wB
  }

  // Calculate the forward propagator: q1=q_i(r) 
  for (r=0; r<M; r++) q1[1][r]=hA[r];
  for (i=1; i<NA; i++) step(q1[i],q1[i+1],g,hA);
  for (i=NA; i<N; i++) step(q1[i],q1[i+1],g,hB);

  // Calculate the backward propagator: q2=q_i^dagger(r) 
  for (r=0; r<M; r++) q2[N][r]=hB[r];
  for (i=N; i>NA+1; i--) step(q2[i],q2[i-1],g,hB);
  for (i=NA+1; i>1; i--) step(q2[i],q2[i-1],g,hA);

  // Calculate concentrations: phi-[r] and phi+[r] 
  Q = 0.0;
  for (r=0; r<M; r++) {
    Q += q1[N][r];
    for (i=1, phiA=0.0; i<=NA; i++) phiA += q1[i][r]*q2[i][r];
    for (i=NA+1, phiB=0.0; i<=N; i++) phiB += q1[i][r]*q2[i][r];
    w[r+2*M] = (phiA/hA[r]-phiB/hB[r]);
    w[r+3*M] = (phiA/hA[r]+phiB/hB[r]);
  }
  Q /= M;
  *lnQ = log(Q);
  for (r=2*M; r<4*M; r++) w[r] /= N*Q;
}
//============================================================
// Solves field equations for w-[r] and w+[r] 
//------------------------------------------------------------
int anderson(double *w, int maxIter, double errTol)
{
  double lambda, err=1.0, S1, S2, U[DIM*DIM], V[DIM], C[DIM];
  double Dh[DIM][M], wh[DIM][M], DD[DIM][DIM], lnQ;
  int    k, m, n, r, nh;

  for (k=1; k<maxIter && err>errTol; k++) {
    diblock(w,&lnQ);
    for (r=0; r<M; r++) Dh[0][r] = w[r+3*M]-1.0;
    S1 = S2 = 0.0;
    for (r=0; r<M; r++) {
      wh[0][r] = w[r+M];
      S1 += Dh[0][r]*Dh[0][r];
      S2 += Dh[0][r];
    }
    err = sqrt(S1/M);
    lambda = 1.0-pow(0.9,double(k));
    nh = (k<DIM)?k-1:DIM-1; // number of histories
    for (m=0; m<=nh; m++) 
      for (r=0, DD[0][m]=0.0; r<M; r++) DD[0][m] += Dh[0][r]*Dh[m][r];

    if (k<2) { // simple mixing
      for (r=0; r<M; r++) w[r+M] += lambda*Dh[0][r];
    } else {   // Anderson mixing
      for (m=1; m<=nh; m++) {
        V[m] = DD[0][0]-DD[0][m];
        for (n=1; n<=m; n++) 
        U[m*nh+n] = U[n*nh+m] = DD[0][0]-DD[0][m]-DD[0][n]+DD[n][m];
      } 
        sim_eq(U,V,C,nh);  // solves for C[n]
        for (r=0; r<M; r++) {
          w[r+M] += lambda*Dh[0][r];
          for (n=1; n<=nh; n++) 
            w[r+M] += C[n]*((wh[n][r]+lambda*Dh[n][r])- 
                          (wh[0][r]+lambda*Dh[0][r])); 
      }
    }

    // copy current field and deviation to history position n 
    n=1+(k-1)%(DIM-1);
    for (r=0; r<M; r++) {
      Dh[n][r] = Dh[0][r];
      wh[n][r] = wh[0][r];
    }
    DD[n][n] = DD[0][0];
    for (m=1; m<n; m++) DD[m][n] = DD[0][m];
    for (m=n+1; m<=nh; m++) DD[n][m] = DD[0][m];
  }

  return k;
}
//============================================================
// Performs Langevin step using predictor-corrector algorithm 
//------------------------------------------------------------
void langevin(double* w, double chi_b, double sigma, double dt)
{
    double w_copy[M], noise[M], lambda1[M], lambda2;
    int    r;
   
    // predictor step 
    for (r=0; r<M; r++) {
      w_copy[r] = w[r];
      noise[r] = normal_dist(0.0,sigma);
      lambda1[r] = w[r+2*M]+2.0*w[r]/chi_b;
      w[r] += -lambda1[r]*dt+noise[r];
    }
    anderson(w,100,1e-4);

    // corrector step 
    for (r=0; r<M; r++) {
      lambda2 = w[r+2*M]+2.0*w[r]/chi_b;
      w[r] = w_copy[r] - 0.5*(lambda1[r]+lambda2)*dt+noise[r];
    }
    anderson(w,100,1e-4);
    
    return;
}
//============================================================
// Output the spherical average of the structure function: S(k) 
//------------------------------------------------------------
void outputS(double* S, double* K, int* wt)
{
    double Save;
    int    k, perm[Mk], flag, temp, count;
    FILE   *out;

    // sort the wavevectors from smallest to largest   
    for (k=0; k<Mk; k++) perm[k] = k;
    do {
      flag = 0;
      for (k=1; k<Mk; k++) 
        if (K[perm[k-1]]>K[perm[k]]) {
          temp = perm[k-1];
          perm[k-1] = perm[k];
          perm[k] = temp;
          flag=1;
        }
    } while (flag==1); 

    // average S(k) over k of equal magnitude and print results
    out = fopen("struct","w");
    Save = 0.0;
    count = 0;
    for (k=0; k<Mk; k++) {
      Save += wt[perm[k]]*S[perm[k]];
      count += wt[perm[k]];
      if ((k==Mk-1)||(K[perm[k+1]]>K[perm[k]]+1e-5)) {
        Save /= count; 
        fprintf(out,"%lf %lf\n",K[perm[k]],Save);
        Save = 0.0;
        count = 0;
      }
    }
    fclose(out);
    
    return;
}
//============================================================
// Key variables in main:
//
// L[i] = size of the simulation box in the i'th dimension
// V = L[0]*L[1]*L[2] = volume of the simulation box
// w[4*M] = array containing N*w-, N*w+, phi-, phi+
// r = (x*m[0]+y)*m[1]+z = array position for (x,y,z)
// N = total number of monomers 
// chi_b = bare chi times N
// C = sqrt(Nbar) = dimensionless concentration
// sigma = standard deviation of the random noise times N
// dt = size of the Langevin step times N
// Hf = free energy per molecule in units of nkT
// lnQ = log of the single-chain partition function
// S[k] = structure function
// K[k] = modulus of the wavevector
// wt[k] = weighting of the wavevectors (equals 1 or 2)
// equil_its = number of equilibration steps
// sim_its = number of simulation steps
// sample_freq = frequency that observables are sampled
//------------------------------------------------------------
int main ()
{
  double coef, chi_b, L[3], Hf, lnQ, V, dt, C, sigma, *w;
  double kx_sq, ky_sq, kz_sq, *K, *S;
  int    r, k, k0, k1, k2, K0, K1, N; 
  int    it, equil_its, sim_its, sample_freq, *wt;
  complex<double> *wk;
  FILE   *in, *out;

  in = fopen("input","r");
  fscanf(in,"%d %d %lf %lf %lf",&N,&NA,&chi_b,&C,&dt);
  fscanf(in,"%d %d %d %lf %lf %lf",&m[0],&m[1],&m[2],&L[0],&L[1],&L[2]);
  fscanf(in,"%d %d %d",&equil_its,&sim_its,&sample_freq);

  sim_its = (sim_its/sample_freq)*sample_freq;  
  M = m[0]*m[1]*m[2];
  Mk = m[0]*m[1]*(m[2]/2+1);
  V = L[0]*L[1]*L[2];
  sigma = sqrt(2.0*M*dt/(C*V));
  NB = N-NA;

  qr = new double[M];
  qk = new complex<double>[M];
  qr_to_qk = fftw_plan_dft_r2c(3,m,qr, reinterpret_cast < fftw_complex * >(qk), FFTW_PATIENT);
  qk_to_qr = fftw_plan_dft_c2r(3,m,reinterpret_cast < fftw_complex * >(qk), qr, FFTW_PATIENT);

  // determine K(k), g(k) and wt(k) for each grid point in Fourier space
  K = new double[Mk];
  g = new double[Mk];
  wt = new int[Mk];
  for (k=0; k<Mk; k++) wt[k]=2;
  for (k0=-(m[0]-1)/2; k0<=m[0]/2; k0++) {
    K0 = (k0<0)?(k0+m[0]):k0;
    kx_sq = k0*k0/(L[0]*L[0]);
    for (k1=-(m[1]-1)/2; k1<=m[1]/2; k1++) {
      K1 = (k1<0)?(k1+m[1]):k1;
      ky_sq = k1*k1/(L[1]*L[1]);
      for (k2=0; k2<=m[2]/2; k2++) {
        kz_sq = k2*k2/(L[2]*L[2]);
        k = k2 + (m[2]/2+1)*(K1+m[1]*K0);
        K[k] = two_pi*sqrt(kx_sq+ky_sq+kz_sq); 
        g[k] = exp(-K[k]*K[k]/(6*N))/M; 
        if ((k2==0)||(k2==m[2]/2)) wt[k]=1;
      }
    }
  }

  // read fields from input file 
  w = new double[4*M];
  for (r=0; r<2*M; r++) fscanf(in,"%lf",&w[r]);
  fclose(in);
  anderson(w,200,1e-4);  // solve for pressure field

  // setup for calculation of structure function
  S = new double[Mk];
  for (k=0; k<Mk; k++) S[k] = 0.0;
  wk = new complex<double>[M];
  wr_to_wk = fftw_plan_dft_r2c(3,m,w, reinterpret_cast < fftw_complex * >(wk), FFTW_PATIENT);

  // set seed for random number generator
  unsigned long int iseed;
  iseed = time(NULL);
  init_genrand(iseed);

  // equilibrate the system
  for (it=1; it<=equil_its; it++) langevin(w,chi_b,sigma,dt);

  // simulate the system and collect statistics for structure function
  for (it=1; it<=sim_its; it++) {
    langevin(w,chi_b,sigma,dt);
    if (it%sample_freq==0) {
      printf("step %5d\n",it);
      fftw_execute(wr_to_wk);
      for (k=0; k<Mk; k++) S[k] += norm(wk[k]);
    }
  }

  // calculate structure function and write to file
  coef = C*V*sample_freq/(chi_b*chi_b*M*M*sim_its);
  for (k=0; k<Mk; k++) S[k] = coef*S[k]-0.5/chi_b;
  outputS(S,K,wt);

  // calculate free energy
  diblock(w,&lnQ);
  for (r=0, Hf=-lnQ; r<M; r++) Hf += (w[r]*w[r]/chi_b-w[r+M])/M;
  printf("free energy: %lf\n",Hf);

  // output the final configuration
  out = fopen("output","w");
  fprintf(out,"%d %d %lf %lf %lf\n",N,NA,chi_b,C,dt);
  fprintf(out,"%d %d %d %lf %lf %lf\n",m[0],m[1],m[2],L[0],L[1],L[2]);
  fprintf(out,"%d %d %d\n",equil_its,sim_its,sample_freq);
  for (r=0; r<2*M; r++) fprintf(out,"%lf\n",w[r]);
  fclose(out);
}
