// fts.cu
//------------------------------------------------------------
// GPU version of the FTS code for a diblock copolymer melt
// Note that lengths are expressed in units of R0=a*N^0.5
//------------------------------------------------------------

#include<math.h>        // math subroutines
#include<stdlib.h>      // standard library
#include<time.h>        // required to seed random number generator
#include<complex>       // for complex-valued variables
#include <cuda.h>       // required for GPUs
#include <cuda_runtime_api.h>   // required for GPUs
#include <cufft.h>      // required for fast Fourier transforms
#include"random.h"      // generates Gaussian random numbers 
#include"sim_eq.h"      // solves simultaneous linear equations
#include "GPUerror.h"   // GPU error handling kernels
#include "GPUkernels.h" // GPU kernels
#define DIM 21          // max histories in Anderson mixing = DIM-1
using namespace std;

//------------------------------------------------------------
// Global CPU variables:
//
// m[i] = number of mesh points in i'th dimension
// M = m[0]*m[1]*m[2] = total number of mesh points
// Mk = m[0]*m[1]*(m[2]/2+1) = total mesh points in k-space
// NA and NB = number of monomers in A and B blocks
// g[k] = FT of Boltzmann weight for bond potential divided by M 

int    m[3], M, Mk, NA, NB;
double two_pi=8*atan(1.0), *g;

//------------------------------------------------------------
// Global GPU variables:
//
// qr[r] = q_i(r) interspliced with q^\dagger_i(r) 
// q1[i] = pointers to q_i(r) 
// q2[i] = pointers to q^\dagger_{N+1-i}(r) 
// qk[k] = Fourier transform (FT) of qr[r] times M/V
// w_gpu[4*M] = GPU copy of w[4*M] in main
// q1_N[r] = q_N(r)
// h[2*M] = stores Boltzmann weights hA[r] and hB[r] sequentially
// g_gpu[k] = GPU copy of g[k]
// qr_to_qk = plan for forward transform from qr[r] to qk[k] 
// qk_to_qr = plan for forward transform from qk[r] to qr[k] 
// wr_to_wk = plan for forward transform from w-[r] to wk[k] 

double **q1, **q2, *qr, *w_gpu, *q1_N, *h, *g_gpu;
cufftDoubleComplex *qk;
cufftHandle wr_to_wk, qr_to_qk, qk_to_qr; 

//============================================================
// declares g_GPU[k], qk[k] and FFT plans for step()
//------------------------------------------------------------
void step_constructor()
{
    HANDLE_ERROR(cudaMalloc((void**)&g_gpu,2*Mk*sizeof(double)));
    HANDLE_ERROR(cudaMemcpy(g_gpu,g,Mk*sizeof(double),cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMemcpy(g_gpu+Mk,g,Mk*sizeof(double),cudaMemcpyHostToDevice));
    HANDLE_ERROR(cudaMalloc((void**)&qk,2*Mk*sizeof(cufftDoubleComplex)));
    HANDLE_ERROR(cufftPlanMany(&qr_to_qk,3,m,NULL,1,0,NULL,1,0,CUFFT_D2Z,2));
    HANDLE_ERROR(cufftPlanMany(&qk_to_qr,3,m,NULL,1,0,NULL,1,0,CUFFT_Z2D,2));
}
//============================================================
// steps propagators: q_in = q_i(r) and q^\dagger_{N+1-i}(r) 
//                    q_out = q_{i+1}(r) and q^\dagger_{N-i}(r) 
//------------------------------------------------------------
void step(double* q_in, double* q_out, int i)
{
    // forward FFT q_i(r) and q^\dagger_{N+1-i}(r)
    HANDLE_ERROR(cufftExecD2Z(qr_to_qk, q_in, qk));

    // multiple g[k]	
    Mult_self << <(2*Mk+TpB-1)/TpB, TpB >> > (qk, g_gpu, 2*Mk);

    // inverse FFT to obtain convolution integrals
    HANDLE_ERROR(cufftExecZ2D(qk_to_qr, qk, q_out));

    // multiple by hA[r] or hB[r] to obtain q_{i+1}(r)
    if (i<NA) Mult_self << <(M+TpB-1)/TpB,TpB >> > (q_out,h,M);
    else Mult_self << <(M+TpB-1)/TpB,TpB >> > (q_out,h+M,M);

    // multiple by hA[r] or hB[r] to obtain q^dagger_{N+1-i}(r) 
    if (i < NB) Mult_self << <(M+TpB-1)/TpB,TpB >> > (q_out+M,h+M,M);
    else Mult_self << <(M+TpB-1)/TpB,TpB >> > (q_out+M,h,M);
}
//==============================================================
// declares the necessary variables and memory for diblock()
//--------------------------------------------------------------
void diblock_constructor()
{
    int i, N=NA+NB;

    HANDLE_ERROR(cudaMalloc((void**)&h,2*M*sizeof(double)));
    HANDLE_ERROR(cudaMalloc((void**)&qr,2*(N+1)*M*sizeof(double)));
    HANDLE_ERROR(cudaMalloc((void**)&w_gpu,4*M*sizeof(double)));
    q1_N = new double[M];
    q1 = new double* [N+1];
    q2 = new double* [N+1];

    // assign pointers to q_i(r) and q^\dagger_i(r) 
    for (i=0; i<=N; i++) {
        q1[i] = qr + 2*i*M;
        q2[(N+1-i)%(N+1)] = qr + (2*i+1)*M;
    }
}
//============================================================
// evaluates phi- and phi+ stored at w+2*M and w+3*M
//------------------------------------------------------------
void diblock(double *w, double *lnQ)
{
    double Q, *phiA_gpu=w_gpu+2*M, *phiB_gpu=w_gpu+3*M;
    int    r, i, N=NA+NB;

    // copy w- and w+ to GPU
    HANDLE_ERROR(cudaMemcpy(w_gpu,w,2*M*sizeof(double),cudaMemcpyHostToDevice));

    // prepare hA[r] and hB[r] on GPU //
    prepare_h << <(2*M+TpB-1)/TpB, TpB >> > (h,w_gpu,N,M);

    // initialize q1 with hA and q2 with hB
    Array_copy << <(2*M+TpB-1)/TpB, TpB >> > (q1[1],h,2*M);

    // iterate the propagators q1 and q2 (stored sequentially)
    for (i=1; i<N; i++) step(q1[i], q1[i+1], i);

    // calculate single-chain partition function
    HANDLE_ERROR(cudaMemcpy(q1_N,q1[N],M*sizeof(double),cudaMemcpyDeviceToHost));
    for (r=0, Q=0.0; r<M; r++) Q += q1_N[r];
    Q /= M;
    *lnQ = log(Q);

    // calculate concentrations
    Array_init << <(2*M+TpB-1)/TpB, TpB >> > (phiA_gpu, 0.0, 2*M);
    for (i=1; i<=NA; i++) 
      sum_phi << <(M+TpB-1)/TpB, TpB >> > (phiA_gpu, q1[i], q2[i], M);
    for (i=NA+1; i<=N; i++) 
      sum_phi << <(M+TpB-1)/TpB, TpB >> > (phiB_gpu, q1[i], q2[i], M);
    normalize_phi << <(M+TpB-1)/TpB, TpB >> > (w_gpu, h, Q, N, M);

    // copy concentrations from GPU to CPU
    HANDLE_ERROR(cudaMemcpy(w+2*M,w_gpu+2*M,2*M*sizeof(double),cudaMemcpyDeviceToHost));
}
//============================================================
// Solves field equations for w-[r] and w+[r]
//------------------------------------------------------------
int anderson(double *w, int maxIter, double errTol)
{
  double lambda, err=1.0, S1, S2, U[DIM*DIM], V[DIM], C[DIM];
  double Dh[DIM][M], wh[DIM][M], DD[DIM][DIM], lnQ;
  int    k, m, n, r, nh;

  for (k=1; k<maxIter && err>errTol; k++) {
    diblock(w,&lnQ);
    for (r=0; r<M; r++) Dh[0][r] = w[r+3*M]-1.0;
    S1 = S2 = 0.0;
    for (r=0; r<M; r++) {
      wh[0][r] = w[r+M];
      S1 += Dh[0][r]*Dh[0][r];
      S2 += Dh[0][r];
    }
    err = pow(S1/M,0.5);

    lambda = 1.0-pow(0.9,double(k));
    nh = (k<DIM)?k-1:DIM-1; // number of histories
    for (m=0; m<=nh; m++)
      for (r=0, DD[0][m]=0.0; r<M; r++) DD[0][m] += Dh[0][r]*Dh[m][r];

    if (k<2) { // simple mixing
      for (r=0; r<M; r++) w[r+M] += lambda*Dh[0][r];
    } else {   // Anderson mixing
      for (m=1; m<=nh; m++) {
        V[m] = DD[0][0]-DD[0][m];
        for (n=1; n<=m; n++)
        U[m*nh+n] = U[n*nh+m] = DD[0][0]-DD[0][m]-DD[0][n]+DD[n][m];
      }
        sim_eq(U,V,C,nh);  // solves for C[n]
        for (r=0; r<M; r++) {
          w[r+M] += lambda*Dh[0][r];
          for (n=1; n<=nh; n++)
            w[r+M] += C[n]*((wh[n][r]+lambda*Dh[n][r])-
                          (wh[0][r]+lambda*Dh[0][r]));
      }
    }

    // copy current field and deviation to history position n
    n=1+(k-1)%(DIM-1);
    for (r=0; r<M; r++) {
      Dh[n][r] = Dh[0][r];
      wh[n][r] = wh[0][r];
    }
    DD[n][n] = DD[0][0];
    for (m=1; m<n; m++) DD[m][n] = DD[0][m];
    for (m=n+1; m<=nh; m++) DD[n][m] = DD[0][m];
  }

  return k;
}
//============================================================
// Performs Langevin step using predictor-corrector algorithm 
//------------------------------------------------------------
void langevin(double* w, double chi_b, double sigma, double dt)
{
    double w_copy[M], noise[M], lambda1[M], lambda2;
    int    r;
   
    // predictor step 
    for (r=0; r<M; r++) {
      w_copy[r] = w[r];
      noise[r] = normal_dist(0.0,sigma);
      lambda1[r] = w[r+2*M]+2.0*w[r]/chi_b;
      w[r] += -lambda1[r]*dt+noise[r];
    }
    anderson(w,100,1e-4);

    // corrector step 
    for (r=0; r<M; r++) {
      lambda2 = w[r+2*M]+2.0*w[r]/chi_b;
      w[r] = w_copy[r] - 0.5*(lambda1[r]+lambda2)*dt+noise[r];
    }
    anderson(w,100,1e-4);
    
    return;
}
//============================================================
// Output the spherical average of the structure function: S(k) 
//------------------------------------------------------------
void outputS(double* S, double* K, int* wt)
{
    double Save;
    int    k, perm[Mk], flag, temp, count;
    FILE   *out;

    // sort the wavevectors from smallest to largest   
    for (k=0; k<Mk; k++) perm[k] = k;
    do {
      flag = 0;
      for (k=1; k<Mk; k++) 
        if (K[perm[k-1]]>K[perm[k]]) {
          temp = perm[k-1];
          perm[k-1] = perm[k];
          perm[k] = temp;
          flag=1;
        }
    } while (flag==1); 

    // average S(k) over k of equal magnitude and print out results
    out = fopen("struct","w");
    Save = 0.0;
    count = 0;
    for (k=0; k<Mk; k++) {
      Save += wt[perm[k]]*S[perm[k]];
      count += wt[perm[k]];
      if ((k==Mk-1)||(K[perm[k+1]]>K[perm[k]]+1e-5)) {
        Save /= count; 
        fprintf(out,"%lf %lf\n",K[perm[k]],Save);
        Save = 0.0;
        count = 0;
      }
    }
    fclose(out);
    
    return;
}
//============================================================
// Key variables in main:
//
// L[i] = size of the simulation box in the i'th dimension
// V = L[0]*L[1]*L[2] = volume of the simulation box
// w[4*M] = array containing N*w-, N*w+, phi-, phi+
// r = (x*m[1]+y)*m[2]+z = array position for (x,y,z)
// N = total number of monomers 
// chi_b = bare chi times N
// C = sqrt(Nbar) = dimensionless concentration
// sigma = standard deviation of the random noise times N
// dt = size of the Langevin step times N
// Hf = Hamiltonian in units of nkT
// lnQ = log of the single-chain partition function
// S[k] = structure function
// K[k] = modulus of the wavevector
// wt[k] = weighting of the wavevectors (equals 1 or 2)
// equil_its = number of equilibration steps
// sim_its = number of simulation steps
// sample_freq = frequency that observables are sampled
// wk[k] = fourier transform of w
// wk_gpu[k] = GPU copy of wk[k]

//------------------------------------------------------------
int main ()
{
  double coef, chi_b, L[3], Hf, lnQ, V, dt, C, sigma, *w;
  double kx_sq, ky_sq, kz_sq, *K, *S;
  int    r, k, k0, k1, k2, K0, K1, N; 
  int    it, equil_its, sim_its, sample_freq, *wt;
  complex<double> *wk;
  cufftDoubleComplex* wk_gpu;
  FILE *in, *out;

  in = fopen("input","r");
  fscanf(in,"%d %d %lf %lf %lf",&N,&NA,&chi_b,&C,&dt);
  fscanf(in,"%d %d %d %lf %lf %lf",&m[0],&m[1],&m[2],&L[0],&L[1],&L[2]);
  fscanf(in,"%d %d %d",&equil_its,&sim_its,&sample_freq);

  sim_its = (sim_its/sample_freq)*sample_freq;  
  M = m[0]*m[1]*m[2];
  Mk = m[0]*m[1]*(m[2]/2+1);
  V = L[0]*L[1]*L[2];
  sigma = sqrt(2.0*M*dt/(C*V));
  NB = N-NA;

  // determine K(k), g(k) and wt(k) for each grid point in Fourier space
  K = new double[Mk];
  g = new double[Mk];
  wt = new int[Mk];
  for (k=0; k<Mk; k++) wt[k]=2;
  for (k0=-(m[0]-1)/2; k0<=m[0]/2; k0++) {
    K0 = (k0<0)?(k0+m[0]):k0;
    kx_sq = k0*k0/(L[0]*L[0]);
    for (k1=-(m[1]-1)/2; k1<=m[1]/2; k1++) {
      K1 = (k1<0)?(k1+m[1]):k1;
      ky_sq = k1*k1/(L[1]*L[1]);
      for (k2=0; k2<=m[2]/2; k2++) {
        kz_sq = k2*k2/(L[2]*L[2]);
        k = k2 + (m[2]/2+1)*(K1+m[1]*K0);
        K[k] = two_pi*pow(kx_sq+ky_sq+kz_sq,0.5); 
        g[k] = exp(-K[k]*K[k]/(6*N))/M; 
        if ((k2==0)||(k2==m[2]/2)) wt[k]=1;
      }
    }
  }

  // set up GPU variables for diblock() and step() //
  diblock_constructor();
  step_constructor();

  // read fields from input file 
  w = new double[4*M];
  for (r=0; r<2*M; r++) fscanf(in,"%lf",&w[r]);
  fclose(in);
  anderson(w,200,1e-4);  // solve for the pressure field

  // setup for calculation of the structure function
  S = new double[Mk];
  for (k=0; k<Mk; k++) S[k] = 0.0;
  wk = new complex<double>[Mk];
  HANDLE_ERROR(cudaMalloc((void**)&wk_gpu, Mk * sizeof(cufftDoubleComplex)));
  HANDLE_ERROR(cufftPlan3d(&wr_to_wk, m[0], m[1], m[2], CUFFT_D2Z));

  // set seed for random number generator
  unsigned long int iseed;
  iseed = time(NULL);
  init_genrand(iseed);

  time_t t_start, t_end;
  // equilibrate the system
  t_start = time(NULL);
  for (it=1; it<=equil_its; it++) langevin(w,chi_b,sigma,dt);
  t_end = time(NULL);
  printf("equlibration time = %ld secs\n", t_end - t_start);


  // simulate the system and collect statistics for structure function
  t_start = time(NULL);
  for (it=1; it<=sim_its; it++) {
    langevin(w,chi_b,sigma,dt);
    if (it%sample_freq==0) {
      printf("step %5d\n",it);
      HANDLE_ERROR(cufftExecD2Z(wr_to_wk, w_gpu, wk_gpu));
      HANDLE_ERROR(cudaMemcpy(wk, wk_gpu, Mk * sizeof(cufftDoubleComplex), cudaMemcpyDeviceToHost));
      for (k=0; k<Mk; k++) S[k] += norm(wk[k]);
    }
  }
  t_end = time(NULL);
  printf("statistics time = %ld secs\n", t_end - t_start);

  // calculate structure function and write to file
  coef = C*V*sample_freq/(chi_b*chi_b*M*M*sim_its);
  for (k=0; k<Mk; k++) S[k] = coef*S[k]-0.5/chi_b;
  outputS(S,K,wt);

  // calculate Hamiltonian
  diblock(w,&lnQ);
  for (r=0, Hf=-lnQ; r<M; r++) Hf += (w[r]*w[r]/chi_b-w[r+M])/M;
  printf("%lf %lf\n",chi_b,Hf);

  // output the final configuration
  out = fopen("output","w");
  fprintf(out,"%d %d %lf %lf %lf\n",N,NA,chi_b,C,dt);
  fprintf(out,"%d %d %d %lf %lf %lf\n",m[0],m[1],m[2],L[0],L[1],L[2]);
  fprintf(out,"%d %d %d\n",equil_its,sim_its,sample_freq);
  for (r=0; r<2*M; r++) fprintf(out,"%lf\n",w[r]);
  fclose(out);

}
